use std::env;

use axum::{
    Router,
    extract::Json,
    response,
    routing::{get, post},
};
use battlesnake::types::{
    Direction,
    simulation::Board,
    wire::{Request, Response},
};
use float_ord::FloatOrd;
use log::{debug, error, info, trace, warn};
use rand::{prelude::*, rng};
use serde::Serialize;
use tokio::{
    net::TcpListener,
    time::{Duration, Instant},
};

#[tokio::main]
async fn main() {
    env_logger::init();

    debug!("Creating routes");
    let app = Router::new()
        .route("/", get(info))
        .route("/start", post(start))
        .route("/move", post(get_move))
        .route("/end", post(end));

    debug!("Creating listener");
    let port = env::var("PORT").unwrap_or_else(|_| "8000".into());
    let listener = TcpListener::bind(format!("0.0.0.0:{port}")).await.unwrap();
    debug!("Starting server");
    axum::serve(listener, app).await.unwrap();
}

async fn info() -> response::Json<Info> {
    info!("got info request");
    response::Json(Info {
        apiversion: "1",
        author: "der-informatiker",
        color: "#00FFEE",
        head: "smart-caterpillar",
        tail: "mouse",
        version: env!("CARGO_PKG_VERSION"),
    })
}

#[derive(Debug, Clone, Serialize)]
struct Info {
    apiversion: &'static str,
    author: &'static str,
    color: &'static str,
    head: &'static str,
    tail: &'static str,
    version: &'static str,
}

async fn start(request: Json<Request>) {
    let board = Board::from(&*request);
    info!("got start request: {board}");
}

#[allow(clippy::too_many_lines)]
async fn get_move(request: Json<Request>) -> response::Json<Response> {
    let start = Instant::now();
    let board = Board::from(&*request);
    let id = board.get_id(&request.you.id).unwrap_or_else(|| {
        error!("My id is not in the simulation board");
        0
    });
    debug!("got move request: {board}");
    let actions = board.valid_actions(id).collect::<Vec<_>>();
    if actions.len() <= 1 {
        info!(
            "only one possible action. Fast forwarding {:?}",
            actions.first()
        );
        return response::Json(Response {
            direction: actions.first().copied().unwrap_or(Direction::Up),
            shout: None,
        });
    }
    info!("valid actions: {actions:?}");

    tokio::task::spawn_blocking(move || {
        let base_turns = board.turn();
        let end_condition: &dyn Fn(&Board) -> Option<_> = match &*request.game.ruleset.name {
            "solo" => &|board| {
                if board.num_snakes() == 0
                    || board.turn() > base_turns + (u32::from(request.you.length) * 3).min(32)
                {
                    Some(())
                } else {
                    None
                }
            },
            _ => &|board| {
                if board.num_snakes() <= 1
                    || board.turn() > base_turns + (u32::from(request.you.length) * 3).min(32)
                {
                    Some(())
                } else {
                    None
                }
            },
        };
        let start_snakes = u32::try_from(board.num_snakes()).unwrap_or(0);
        let score_fn: &dyn Fn(&Board, u8) -> u32 = match &*request.game.ruleset.name {
            "solo" => &|board, _| board.turn(),
            _ => &|board, id| {
                if board.alive(id) {
                    1 + start_snakes - u32::try_from(board.num_snakes()).unwrap_or(start_snakes)
                } else {
                    0
                }
            },
        };

        let mut mcts_managers: Vec<_> = (0..request.board.snakes.len())
            .map(|id| MctsManager::new(u8::try_from(id).unwrap()))
            .collect();
        let c = f32::sqrt(2.0);
        while start.elapsed() < Duration::from_millis(250) {
            let mut board = board.clone();
            while end_condition(&board).is_none() {
                let actions: Vec<_> = mcts_managers
                    .iter_mut()
                    .filter_map(|mcts_manager| {
                        mcts_manager
                            .next_action(&board, c)
                            .map(|action| (mcts_manager.snake, action))
                    })
                    .collect();
                board.next_turn(&actions);
                if actions.is_empty() {
                    break;
                }
            }
            board.simulate_random(end_condition);
            for mcts_manager in &mut mcts_managers {
                let id = mcts_manager.snake;
                let score = score_fn(&board, id);
                mcts_manager.apply_score(score);
            }
        }
        let my_mcts_manager = &mcts_managers[usize::from(id)];
        for action in actions {
            let score = my_mcts_manager.base.next[usize::from(action)]
                .as_ref()
                .map(|info| info.score as f32 / info.played as f32);
            if let Some(score) = score {
                info!("{action:?} -> {score}");
            } else {
                info!("{action:?} -> None");
            }
        }

        let action = my_mcts_manager
            .base
            .next
            .iter()
            .enumerate()
            .filter_map(|(index, info)| {
                info.as_ref().map(|info| {
                    (
                        match index {
                            0 => Direction::Up,
                            1 => Direction::Down,
                            2 => Direction::Left,
                            3 => Direction::Right,
                            _ => unreachable!(),
                        },
                        info,
                    )
                })
            })
            .max_by_key(|(_, info)| FloatOrd(info.score as f32 / info.played as f32))
            .map(|(action, _)| action);

        if let Some(action) = action {
            info!(
                "found action {action:?} after {} simulations.",
                my_mcts_manager.base.played
            );
        } else {
            warn!("unable to find a valid action");
        }
        info!("chose {action:?}");
        response::Json(Response {
            direction: action.unwrap_or(Direction::Up),
            shout: None,
        })
    })
    .await
    .unwrap()
}

async fn end(request: Json<Request>) {
    let board = Board::from(&*request);
    info!("got end request: {board}");
}

#[derive(Debug)]
struct ActionInfo {
    score: u32,
    played: u32,
    next: Box<[Option<ActionInfo>; 4]>,
}

impl ActionInfo {
    fn new() -> Self {
        Self {
            score: 0,
            played: 0,
            next: Box::new([None, None, None, None]),
        }
    }

    fn uct(&self, c: f32) -> [Option<f32>; 4] {
        let mut ucts = [None; 4];
        for (action, uct) in self.next.iter().zip(ucts.iter_mut()) {
            if let Some(action) = action {
                let exploitation = action.score as f32 / action.played as f32;
                let exploration = f32::sqrt(f32::ln(self.played as f32) / action.played as f32);
                uct.replace(c.mul_add(exploration, exploitation));
            }
        }
        ucts
    }
}

#[derive(Debug)]
struct MctsManager {
    base: ActionInfo,
    actions: Vec<Direction>,
    expanded: bool,
    snake: u8,
}

impl MctsManager {
    fn new(snake: u8) -> Self {
        Self {
            base: ActionInfo::new(),
            actions: Vec::new(),
            expanded: false,
            snake,
        }
    }

    fn apply_score(&mut self, score: u32) {
        self.base.played += 1;
        self.base.score += score;
        let mut current = &mut self.base;
        for action in &self.actions {
            let Some(new_current) = &mut current.next[usize::from(*action)] else {
                error!("got action without actioninfo");
                break;
            };
            current = new_current;
            current.played += 1;
            current.score += score;
        }
        self.actions.clear();
        self.expanded = false;
    }

    fn next_action(&mut self, board: &Board, c: f32) -> Option<Direction> {
        if self.expanded {
            return None;
        }

        let mut current = &mut self.base;
        for action in &self.actions {
            let Some(new_current) = &mut current.next[usize::from(*action)] else {
                error!("got action without actioninfo");
                return None;
            };
            current = new_current;
        }

        let ucts = current.uct(c);
        let valid_actions = board.valid_actions(self.snake);
        let ucts: Vec<_> = valid_actions
            .map(|action| (action, ucts[usize::from(action)]))
            .collect();
        trace!("got actions: {ucts:?}");
        if ucts.iter().any(|(_, uct)| uct.is_none()) {
            let action = ucts
                .iter()
                .filter(|(_, uct)| uct.is_none())
                .choose(&mut rng())?
                .0;
            self.expanded = true;
            current.next[usize::from(action)].replace(ActionInfo::new());
            self.actions.push(action);
            return Some(action);
        }

        let action = ucts
            .iter()
            .max_by_key(|(_, uct)| FloatOrd(uct.unwrap_or(f32::NEG_INFINITY)))
            .map(|(action, _)| *action);
        if let Some(action) = action {
            self.actions.push(action);
        }
        action
    }
}
